#ifndef NRF_SVC__
#define NRF_SVC__

#ifdef SVCALL_AS_NORMAL_FUNCTION
#define SVCALL(number, return_type, signature, ...) return_type signature
#else

#ifndef SVCALL
#if defined (__CC_ARM)
#define SVCALL(number, return_type, signature, ...) return_type __svc(number) signature
// *********************************************************
// BSIEVER: Updated for potential clang support
#elif defined (__GNUC__) && !defined(__clang__) 
#define SVCALL0(number, return_type, signature) \
  _Pragma("GCC diagnostic ignored \"-Wreturn-type\"") \
  _Pragma("GCC diagnostic ignored \"-Wunused-function\"") \
  __attribute__((naked)) static return_type signature \
  { \
    __asm( \
        "svc %0\n" \
        "bx r14" : : "I" (number) : "r0" \
    ); \
  }

#define SVCALL(_n, _r, _s,...) SVCALL0(_n, _r, _s)

#elif defined (__clang__) 
// BSIEVER: Clang Compilation Support Macros
#define SVCALL0(number, rettype, signature)         \
__attribute__((always_inline)) static inline rettype signature {  \
    register rettype retVal asm("r0");              \
    asm volatile (                                  \
       "svc %[i]\n\t"                               \
       : "=r" (retVal)                              \
       :[i] "i" (number)                            \
       : "r0");                                     \
    return retVal;                                  \
}

#define SVCALL1(number, rettype, signature, p1 )    \
__attribute__((always_inline)) static inline rettype signature {  \
    register typeof(p1) _p1 asm("r0") = p1;         \
    asm volatile (                                  \
       "svc %[i]\n\t"                               \
       : "=r" (_p1)                                 \
       :[i] "i" (number), [a] "r" (_p1)             \
       : "r0");                                     \
    return (rettype)_p1;                             \
}

#define SVCALL2(number, rettype, signature, p1, p2)  \
__attribute__((always_inline)) static inline rettype signature {  \
    register typeof(p1) _p1 asm("r0") = p1;         \
    register typeof(p2) _p2 asm("r1") = p2;         \
    asm volatile (                                  \
       "svc %[i]\n\t"                               \
       : "=r" (_p1)                                 \
       :[i] "i" (number), [a] "r" (_p1), [b] "r" (_p2) \
       : "r0");                                     \
    return (rettype)_p1;                            \
}
#define SVCALL3(number, rettype, signature, p1, p2, p3)  \
__attribute__((always_inline)) static inline rettype signature {  \
    register typeof(p1) _p1 asm("r0") = p1;         \
    register typeof(p2) _p2 asm("r1") = p2;         \
    register typeof(p3) _p3 asm("r2") = p3;         \
    asm volatile (                                  \
       "svc %[i]\n\t"                               \
       : "=r" (_p1)                                 \
       :[i] "i" (number), [a] "r" (_p1), [b] "r" (_p2), [c] "r" (_p3) \
       : "r0");                                     \
    return (rettype)_p1;                            \
}
#define SVCALL4(number, rettype, signature, p1, p2, p3, p4)  \
__attribute__((always_inline)) static inline rettype signature {  \
    register typeof(p1) _p1 asm("r0") = p1;         \
    register typeof(p2) _p2 asm("r1") = p2;         \
    register typeof(p3) _p3 asm("r2") = p3;         \
    register typeof(p4) _p4 asm("r3") = p4;         \
    asm volatile (                                  \
       "svc %[i]\n\t"                               \
       : "=r" (_p1)                                 \
       :[i] "i" (number), [a] "r" (_p1), [b] "r" (_p2), [c] "r" (_p3), [d] "r" (_p4) \
       : "r0");                                     \
    return (rettype)_p1;                            \
}


// Macro Overloading. Based on: http://stackoverflow.com/questions/11761703/overloading-macro-on-number-of-arguments
#define GET_SVCMACRO(_0, _n, _r, _s, _p1, _p2, _p3, _p4, NAME, ...) NAME
#define SVCALL(...) GET_SVCMACRO(_0, ##__VA_ARGS__, SVCALL4, SVCALL3, SVCALL2, SVCALL1, SVCALL0)(__VA_ARGS__)
// *********************************************************
#elif defined (__ICCARM__)
#define PRAGMA(x) _Pragma(#x)
#define SVCALL(number, return_type, signature) \
PRAGMA(swi_number = number) \
 __swi return_type signature;
#else
#define SVCALL(number, return_type, signature) return_type signature  
#endif
#endif  // SVCALL

#endif  // SVCALL_AS_NORMAL_FUNCTION
#endif  // NRF_SVC__
